module Blufin

    class YmlResourceValidator < Blufin::YmlValidatorBase

        CONFIG          = 'config'
        CONFIG_INTERNAL = 'internal'
        CONFIG_OAUTH    = 'oauth'

        STRUCTURE = [
            {
                :section_name => CONFIG,
                :section_type => Blufin::YmlValidatorBase::SECTION_TYPE_FIXED,
                :section_data => {
                    CONFIG_INTERNAL => {
                        :type         => Blufin::YmlValidatorBase::FIELD_TYPE_NESTED_DATA,
                        :required     => false,
                        :section_data => {
                            'GET'    => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'POST'   => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'PATCH'  => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'DELETE' => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            }
                        }
                    },
                    CONFIG_OAUTH    => {
                        :type         => Blufin::YmlValidatorBase::FIELD_TYPE_NESTED_DATA,
                        :required     => false,
                        :section_data => {
                            'GET'    => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'POST'   => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'PATCH'  => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            },
                            'DELETE' => {
                                :type     => Blufin::YmlValidatorBase::FIELD_TYPE_BLANK,
                                :required => false
                            }
                        }
                    }
                }
            },
            {   # TODO - 12/20/18 - I'm pretty sure this is all wrong...
                :section_name => 'request',
                :section_type => Blufin::YmlValidatorBase::SECTION_TYPE_DYNAMIC,
                :section_data => {
                    :section_alphabetical => false,
                    :section_regex        => /\A(app\.|common\.|config\.|mock\.)*[a-z_]+(\[\]|\[link\])*\z/,
                    :section_fields       => {
                        'type'               => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'flag'               => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'fkey'               => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'default'            => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'encrypted'          => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'description'        => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                        'required_if'        => {
                            :type     => Blufin::YmlValidatorBase::FIELD_TYPE_TEXT,
                            :required => false
                        },
                    }
                }
            }
        ]

        # Initialize the class.
        # @return void
        def initialize(site, error_handler)

            @site      = Blufin::SiteResolver::validate_site(site)
            @site_path = Blufin::SiteResolver::get_site_location(@site)
            @site_name = Blufin::SiteResolver::get_site_name(@site)

            @error_handler = error_handler

            # Create resource folder if it doesn't exist.
            resource_folder = "#{Blufin::SiteResolver::get_site_location(@site)}/yml/api/resources"
            Blufin::Files::create_directory(resource_folder) unless Blufin::Files::path_exists(resource_folder)

            # Loop through the array of files.
            Blufin::Files.get_files_in_dir(resource_folder).each { |file| scan_file(file) }

        end

        private

        # The main function that does the file scanning.
        # @return void
        def scan_file(file)

            return unless file_is_valid(file)

            # Get the YML data from the file.
            begin
                data = YAML.load_file(File.expand_path(file))
            rescue Exception => e
                @error_handler.add_error(Blufin::YmlErrorHandler::FILE_INVALID, file, nil, nil, file)
                return
            end

            # TODO - FINISH/FIX
            # data = validate_single_file(@site, @config_file_app, alter_structure_for_app(STRUCTURE_APP), error_handler)

        end

        # Validates the file.
        # @return List
        def file_is_valid(file)
            valid = true
            # Make sure this is a '.yml.' file.
            unless Blufin::YmlCommon.is_yml_file(file)
                @error_handler.add_error(Blufin::YmlErrorHandler::FILE_INVALID, file, nil, nil, 'Expected .yml file.')
                valid = false
            end
            valid
        end

    end

end