module AppCommand

    class APIMeta < ::Convoy::ActionCommand::Base

        def execute

            begin

                @opts = command_options
                @args = arguments

                opts_validate

                @site          = Blufin::SiteResolver::validate_site(@args[0])
                @error_handler = Blufin::YmlErrorHandler.new(@site)

                Blufin::SiteEmbedded::init(Blufin::ScannerJavaEmbeddedObjects::new(@site, @error_handler).get_data)

                @yml_config = Blufin::YmlConfigValidator.new(@site, @error_handler)

                Blufin::SiteAuth::init(@yml_config.config_data[Blufin::YmlConfigValidator::GLOBAL][Blufin::YmlConfigValidator::GLOBAL_AUTHENTICATION])

                @yml_schema = Blufin::YmlSchemaValidator.new(@site, @error_handler)

                opts_routing

            rescue => e

                Blufin::Terminal::print_exception(e)

            end

        end

        def opts_validate

            # Show help if no flags.
            if Blufin::Routes::flags_set(@opts) == 0

                system("#{App::GEM_NAME} p am -h")
                exit

            end

        end

        def opts_routing

            system('clear') unless @opts[:as_json]

            @yml_schema.get_yml_schema_outputter.show_auto_increment if @opts[:show_auto_increment]
            @yml_schema.get_yml_schema_outputter.show_enums if @opts[:show_enums]
            @yml_schema.get_yml_schema_outputter.show_resources(@opts[:as_json]) if @opts[:show_resources]

        end

    end

end