-- Create client table
CREATE TABLE `client` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `client_name` VARCHAR(64) NOT NULL,
  PRIMARY KEY (`id`),
  UNIQUE INDEX `config_client_client_name_uniq_idx` (`client_name`)
) ENGINE = InnoDB;

-- Create db table
CREATE TABLE `db` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `project_id` INT NULL,
  `db_name` VARCHAR(32) NOT NULL COMMENT 'The name for the database as displayed by our internal systems.',
  `host` VARCHAR(255) NOT NULL COMMENT 'The URL of the server (through SSH connection -- IE: username.abc123.eu-west-1.rds.amazonaws.com ).',
  `user` TEXT NOT NULL COMMENT 'The username used to connect to the server.',
  `password` TEXT NOT NULL COMMENT 'The password used to connect to the server.',
  `environment` VARCHAR(11) NOT NULL COMMENT 'The environment for which this database will apply to.',
  `schema_type` VARCHAR(6) NOT NULL COMMENT 'The schema type (IE: config, common or app).',
  `jdbc_url` VARCHAR(255) NOT NULL COMMENT 'The URL of the server (through Java JDBC connection -- IE: jdbc:mysql://abc123.com:3306 ).',
  PRIMARY KEY (`id`),
  UNIQUE INDEX `config_db_db_name_uniq_idx` (`db_name`),
  INDEX `config_db_environment_idx` (`environment`)
) ENGINE = InnoDB;

-- Create db_configuration table
CREATE TABLE `db_configuration` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `db_configuration_name` VARCHAR(32) NOT NULL COMMENT 'The name for this configuration.',
  `pool_max_size` INT NOT NULL COMMENT 'The maximum size of the connection pool.',
  PRIMARY KEY (`id`),
  UNIQUE INDEX `config_db_configuration_db_configuration_name_uniq_idx` (`db_configuration_name`)
) ENGINE = InnoDB;

-- Create db_configuration_property table
CREATE TABLE `db_configuration_property` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `db_configuration_id` INT NOT NULL,
  `key` VARCHAR(64) NOT NULL COMMENT 'Database property key (IE: charSet)',
  `value` VARCHAR(128) NOT NULL COMMENT 'Database property value (IE: UTF-8)',
  PRIMARY KEY (`id`)
) ENGINE = InnoDB;

-- Create profile table
CREATE TABLE `profile` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `project_id` INT NOT NULL,
  `profile_name` VARCHAR(32) NOT NULL COMMENT 'The name of the profile (IE: localhost, us-west-1, etc.)',
  `environment` VARCHAR(11) NOT NULL COMMENT 'The environment in which the application will run under this profile.',
  `timezone` VARCHAR(64) NOT NULL COMMENT 'The timezone in which associated application will run.',
  `message_queue_host` VARCHAR(255) NOT NULL COMMENT 'The message queue URL.',
  `slack_log_url` VARCHAR(255) NOT NULL COMMENT 'The Slack Log channel webhook URL.',
  `slack_error_url` VARCHAR(255) NOT NULL COMMENT 'The Slack Error channel webhook URL.',
  `authorization_header` TEXT NOT NULL COMMENT 'The Authorization Header used for internal API calls.',
  PRIMARY KEY (`id`),
  INDEX `config_profile_profile_name_idx` (`profile_name`),
  INDEX `config_profile_environment_idx` (`environment`)
) ENGINE = InnoDB;

-- Create profile_api table
CREATE TABLE `profile_api` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `profile_id` INT NOT NULL,
  `common_db_id` INT NOT NULL,
  `common_db_configuration_id` INT NOT NULL,
  PRIMARY KEY (`id`)
) ENGINE = InnoDB;

-- Create profile_cron table
CREATE TABLE `profile_cron` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `profile_id` INT NOT NULL,
  PRIMARY KEY (`id`)
) ENGINE = InnoDB;

-- Create profile_worker table
CREATE TABLE `profile_worker` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `profile_id` INT NOT NULL,
  PRIMARY KEY (`id`)
) ENGINE = InnoDB;

-- Create project table
CREATE TABLE `project` (
  `id` INT NOT NULL AUTO_INCREMENT,
  `client_id` INT NOT NULL,
  `project_name` VARCHAR(64) NOT NULL COMMENT 'Must be all lowercase & hyphen separated (IE: go-eat)',
  `project_name_pascal_case` VARCHAR(64) NOT NULL COMMENT 'Must be camel-cased (IE: GoEat)',
  `project_title` VARCHAR(128) NOT NULL COMMENT 'The way you want the project title/name to be displayed in UIs (IE: Go-EAT)',
  `domain` VARCHAR(128) NOT NULL COMMENT 'The domain where the project will be hosted (IE: com, co.uk, app). This combined with the {project:project_name:Project name} should make up the public URL.',
  `alias` VARCHAR(8) NOT NULL COMMENT 'A short alias for the project, mainly for use in CLI commands (IE: ge)',
  `port_range` VARCHAR(12) NOT NULL COMMENT 'Ports reserved for this project, must be in sections of 20 (IE: 6000-6019)',
  `encryption_key` VARCHAR(64) NOT NULL COMMENT 'Encryption Key for encrypting/decrypting sensitive data. This can never change!',
  `encryption_iv` VARCHAR(64) NOT NULL COMMENT 'Encryption IV for encrypting/decrypting sensitive data. This can never change!',
  PRIMARY KEY (`id`),
  UNIQUE INDEX `config_project_project_name_uniq_idx` (`project_name`),
  UNIQUE INDEX `config_project_alias_uniq_idx` (`alias`)
) ENGINE = InnoDB;